"""Topologic functions executed on extended geodataframe."""

from typing import Literal
import geopandas as gpd
import pandas as pd
from shapely.geometry import Point

""" 
In this block we define supporting functions. Supporting functions are ignored
by the logical-validation module. They are used by and shared between topological
validation rules.
"""


def _layers_from_datamodel(layers, datamodel):
    if len(layers) > 1:
        series = pd.concat([getattr(datamodel, i)["geometry"] for i in layers])
    else:
        series = getattr(datamodel, layers[0])["geometry"]
    return series.loc[series.index.notna()]


def _lines_snap_at_boundaries(line, other_line, tolerance):
    snaps_start = any(
        line.boundary.geoms[0].distance(i) < tolerance
        for i in other_line.boundary.geoms
    )
    snaps_end = any(
        line.boundary.geoms[-1].distance(i) < tolerance
        for i in other_line.boundary.geoms
    )
    return snaps_start, snaps_end


def _point_not_overlapping_line(point, line, tolerance):
    if line.boundary:
        snaps_start = line.boundary.geoms[0].distance(point) < tolerance
        snaps_end = line.boundary.geoms[-1].distance(point) < tolerance
    else:
        snaps_start = snaps_end = (
            Point(list(line.coords)[0]).distance(point) < tolerance
        )
    if not any([snaps_start, snaps_end]):
        not_overlapping = line.distance(point) > tolerance
    else:
        not_overlapping = True
    return not_overlapping


def _line_not_overlapping_line(line, other_line, tolerance):
    # check if lines snap at boundaries
    snaps_start, snaps_end = _lines_snap_at_boundaries(line, other_line, tolerance)

    # in the case we have two lines with only and-points and all end-points overlap, these are overlapping lines
    if (
        (snaps_start and snaps_end)
        and (len(line.coords) == 2)
        and (len(other_line.coords) == 2)
    ):
        return False

    # compare the shortest line with the longest line
    # if ends snap, we check all coordinates in between, else we check all.
    if line.length < other_line.length:
        not_overlapping = all(
            (
                _point_not_overlapping_line(Point(i), other_line, tolerance)
                for i in line.coords
            )
        )
    else:
        not_overlapping = all(
            (
                _point_not_overlapping_line(Point(i), line, tolerance)
                for i in other_line.coords
            )
        )

    return not_overlapping


def _not_overlapping_line(row, gdf, sindex, tolerance, exclude_row=True):
    geometry = row["geometry"]
    indices = list(sindex.intersection(geometry.buffer(tolerance).bounds))
    if exclude_row:
        indices = [i for i in indices if i != gdf.index.get_loc(row.name)]
    if indices:
        gdf_select = gdf.iloc[indices]
        if geometry.geom_type == "LineString":
            not_overlapping = all(
                _line_not_overlapping_line(geometry, i, tolerance)
                for i in gdf_select["geometry"]
            )
        elif geometry.geom_type == "Point":
            not_overlapping = all(
                _point_not_overlapping_line(geometry, i, tolerance)
                for i in gdf_select["geometry"]
            )
    else:
        not_overlapping = True
    return not_overlapping


def _not_overlapping_point(row, gdf, sindex, tolerance, exclude_row=True):
    geometry = row["geometry"]
    indices = list(sindex.intersection(geometry.buffer(tolerance).bounds))
    if exclude_row:
        indices = [i for i in indices if i != gdf.index.get_loc(row.name)]
    if indices:
        gdf_select = gdf.iloc[indices]
        not_overlapping = all(
            geometry.distance(i) > tolerance for i in gdf_select["geometry"]
        )
    else:
        not_overlapping = True
    return not_overlapping


def _snap_nodes(row, series, tolerance):
    series_selec = series.loc[~(series.index == row.name)]
    indices = series_selec.loc[
        (series_selec.distance(row["geometry"]) < tolerance)
    ].index.to_list()
    geom = None
    if indices:
        indices.sort()
        if indices[0] < row.name:
            geom = series.loc[indices[0]]
    if geom is None:
        geom = row["geometry"]

    return geom


def _get_nodes(gdf, tolerance):
    # start and end-nodes to GeoSeries
    nodes_series = gdf["geometry"].apply(lambda x: Point(x.coords[0]))
    nodes_series = pd.concat(
        [nodes_series, gdf["geometry"].apply(lambda x: Point(x.coords[-1]))]
    ).reset_index(drop=True)

    # snap nodes within tolerance: nodes within tolerance get the coordinate
    # of the first node.
    nodes_series = gpd.GeoSeries(
        gpd.GeoDataFrame(nodes_series, columns=["geometry"]).apply(
            lambda x: _snap_nodes(x, nodes_series, tolerance), axis=1
        )
    )

    # as all is snapped we can filter unique points
    nodes_series = gpd.GeoSeries(nodes_series.unique())

    return nodes_series


def _only_end_nodes(row, series, sindex, tolerance):
    geometry = row["geometry"]
    indices = list(sindex.intersection(geometry.bounds))
    if indices:
        series_select = series.loc[indices]
        only_end_nodes = all(
            _point_not_overlapping_line(i, geometry, tolerance) for i in series_select
        )
    else:
        only_end_nodes = True

    return only_end_nodes


def _structure_at_intersection(
    geometry, other_geometry, struc_gdf, struc_sindex, tolerance
):
    # compute the intersection and buffer it with tolerance
    intersection = geometry.intersection(other_geometry).buffer(tolerance)

    # get the structures that intersect the intersection bounds
    indices = struc_sindex.intersection(intersection.bounds)

    # check if any of these structures are within intersection
    structure_present = any(i.intersects(intersection) for i in struc_gdf.iloc[indices])

    return structure_present


def _structures_at_intersections(
    row, gdf, sindex, struc_series, struc_sindex, tolerance
):
    geometry = row["geometry"]

    # get a selection of hydroobjects that intersect geometry
    indices = sindex.intersection(geometry.bounds)
    indices = [i for i in indices if i != gdf.index.get_loc(row.name)]
    gdf_select = gdf.iloc[indices]
    gdf_select = gdf_select.loc[gdf_select["geometry"].crosses(geometry)]

    # if not emtpy, check if there is a structure at every intersection
    if not gdf_select.empty:
        structure_at_intersections = (
            gdf_select["geometry"]
            .apply(
                lambda x: _structure_at_intersection(
                    x, geometry, struc_series, struc_sindex, tolerance
                )
            )
            .all()
        )
    else:
        structure_at_intersections = True

    return structure_at_intersections


def _intersects_end_node(geometry, series, sindex, tolerance):
    # get inidices in bounds around geometry
    geometry = geometry.buffer(tolerance)
    indices = list(sindex.intersection(geometry.bounds))
    # see if there are start-nodes within within tolerance
    series_select = series.iloc[indices]
    series_select = series_select.loc[series_select.within(geometry)]

    # if dataframe is emtpy, there are no start nodes intersecting end-node
    if series_select.empty:
        intersects_start_node = False
    else:
        intersects_start_node = True
    return intersects_start_node


def _structure_at_boundary(geometry, series, distance):
    return (series.distance(geometry) < distance).any()


def _structures_at_boundaries(
    row, areas_gdf, areas_sindex, struc_series, struc_sindex, tolerance, distance
):
    geometry = row["geometry"]

    # check if hydroobject crosses boundary
    indices = list(areas_sindex.intersection(geometry.bounds))
    areas_gdf_select = areas_gdf.iloc[indices]
    areas_gdf_select = areas_gdf_select.loc[
        areas_gdf_select["geometry"].boundary.intersects(geometry)
    ]

    if areas_gdf_select.empty:
        structure_at_boundary = True
    else:
        # check if hydroobject intersects structures within tolerance
        indices = list(struc_sindex.intersection(geometry.bounds))
        struc_series_select = struc_series.iloc[indices]
        struc_series_select = struc_series_select.loc[
            struc_series_select.distance(geometry) < tolerance
        ]
        if struc_series_select.empty:
            # if there are boundaries, there should be structures
            structure_at_boundary = False
        else:
            # if there are structures they should be close enough to the boundary
            # create series with intersections
            intersections_series = areas_gdf_select["geometry"].boundary.intersection(
                geometry
            )

            # if there is a structure close enough to all boundaries: True. Else: False
            structure_at_boundary = intersections_series.apply(
                lambda x: _structure_at_boundary(x, struc_series_select, distance)
            ).all()

    return structure_at_boundary


def _distant_to_others(row, gdf, sindex, distance):
    geometry = row["geometry"]

    # get a selection of other objects that intersect with
    indices = sindex.intersection(geometry.buffer(distance).bounds)
    indices = [i for i in indices if i != gdf.index.get_loc(row.name)]
    gdf_select = gdf.iloc[indices]
    distant_to_others = all(
        i.distance(geometry) > distance for i in gdf_select["geometry"]
    )
    return distant_to_others


def _no_struc_on_line(geometry, struc_series, sindex, tolerance):
    indices = list(sindex.intersection(geometry.bounds))
    if indices:
        struc_series_select = struc_series.iloc[indices]
        no_struc_on_line = all(
            _point_not_overlapping_line(i, geometry, tolerance)
            for i in struc_series_select
        )
    else:
        no_struc_on_line = True
    return no_struc_on_line


def _compare_longitudinal(
    row, parameter, compare_gdf, compare_parameter, direction, logical_operator
):
    branch_id = row["branch_id"]
    select_gdf = compare_gdf.loc[compare_gdf["branch_id"] == branch_id]
    right = None
    result = None

    if not select_gdf.empty:
        if direction == "downstream":
            # select downstream objects
            select_gdf = select_gdf[
                select_gdf["branch_offset"].gt(row["branch_offset"])
            ]
            # if any, select value from closest object
            if not select_gdf.empty:
                right = select_gdf.loc[select_gdf["branch_offset"].idxmin()][
                    compare_parameter
                ]
        elif direction == "upstream":
            # select downstream objects
            select_gdf = select_gdf[
                select_gdf["branch_offset"].lt(row["branch_offset"])
            ]
            # if any, select value from closest object
            if not select_gdf.empty:
                right = select_gdf.loc[select_gdf["branch_offset"].idxmax()][
                    compare_parameter
                ]

    # if there is a right, there is something to compare
    if right is not None:
        if logical_operator == "GT":
            result = row[parameter] > right
        elif logical_operator == "LT":
            result = row[parameter] < right
    return result


""" 
In this block we define topologic functions.
"""


def snaps_to_hydroobject(gdf, datamodel, method, tolerance=0.001, dtype=bool):
    """
    Check if geometries snap to HydroObject

    Parameters
    ----------
    gdf : ExtendedGeoDataframe
        ExtendedGeoDataFrame, typically a layer in a HyDAMO datamodel class
    datamodel : HyDAMO
        HyDAMO datamodel class
    method : str, options: 'intersecting', 'overal', 'centroid', 'ends'
        Method that can be used to deterine nearest hydrobject
    tolerance : numeric
        Tolerance used to snap to the hydroobjct
    dtype : dtype, optional
        Dtype to assign to the result series. The default is bool.

    Returns
    -------
    Pandas Series
        Default dtype is bool

    """

    branches = datamodel.hydroobject
    gdf.snap_to_branch(branches, snap_method=method, maxdist=tolerance)
    series = ~gdf.branch_offset.isna()
    return series.astype(dtype)


def geometry_length(
    gdf, datamodel, length, statistic: Literal["min", "max"] = "min", dtype=bool
):
    """Check if geometrie length is longer/shorter than a value

    Parameters
    ----------
    gdf : ExtendedGeoDataframe
        ExtendedGeoDataFrame, typically a layer in a HyDAMO datamodel class
    datamodel : HyDAMO
        HyDAMO datamodel class
    length : numeric
        Length to compare the geometry length to
    statistic : str, options: 'min', 'max'
        Use length as minimal or maximal length. Default is 'min'
    dtype : dtype, optional
        Dtype to assign to the result series. The default is bool.

    Returns
    -------
    Pandas Series
        Default dtype is bool

    """
    if statistic == "min":
        series = gdf["geometry"].length >= length
    elif statistic == "max":
        series = gdf["geometry"].length <= length
    return series.astype(dtype)


def not_overlapping(gdf, datamodel, tolerance):
    """Check if an objects LineString geometry is not overlapping other object
    of the same layer.

    Parameters
    ----------
    gdf : ExtendedGeoDataframe
        ExtendedGeoDataFrame, typically a layer in a HyDAMO datamodel class
    datamodel : HyDAMO
        HyDAMO datamodel class
    tolerance : numeric
        Max tolerance for overlapping

    Returns
    -------
    Pandas Series
        Default dtype is bool

    """
    sindex = gdf.sindex
    if (gdf.geom_type == "LineString").all():
        return gdf.apply(
            lambda x: _not_overlapping_line(x, gdf, sindex, tolerance), axis=1
        )
    elif (gdf.geom_type == "Point").all():
        return gdf.apply(
            lambda x: _not_overlapping_point(x, gdf, sindex, tolerance), axis=1
        )
    else:
        raise TypeError(
            f"GeoDataFrame has invalid geometry types: {gdf.geom_type.unique()}. Implemented for this function: [Point, LineString]"
        )


def splitted_at_junction(gdf, datamodel, tolerance):
    """Check if line is splitted when it can considered to be a junction

    Parameters
    ----------
    gdf : ExtendedGeoDataframe
        ExtendedGeoDataFrame, HyDAMO hydroobject layer
    datamodel : HyDAMO
        HyDAMO datamodel class
    tolerance : numeric
        Max tolerance for junction nodes

    Returns
    -------
    Pandas Series
        Default dtype is bool

    """
    # get the nodes of the hydroobjects within tolerance
    nodes_series = _get_nodes(gdf, tolerance)

    # check for lines if there are nodes on segment outside tolerance of
    # the start-node and end-node.
    sindex = nodes_series.sindex
    return gdf.apply(
        (lambda x: _only_end_nodes(x, nodes_series, sindex, tolerance)), axis=1
    )


def structures_at_intersections(gdf, datamodel, structures, tolerance):
    """Check if there are structures at intersections

    Parameters
    ----------
    gdf : ExtendedGeoDataframe
        ExtendedGeoDataFrame, HyDAMO hydroobject layer
    datamodel : HyDAMO
        HyDAMO datamodel class
    structures: str or list
        HyDAMO structures to be expected at intersections
        ("stuw", "duikersifonhevel") Presented as a string or list
    tolerance : numeric
        Max tolerance for junction nodes

    Returns
    -------
    Pandas Series
        Default dtype is bool

    """
    # make a geodataframe from structures list
    struc_series = _layers_from_datamodel(structures, datamodel)
    # create spatial indices
    sindex = gdf.sindex
    struc_sindex = struc_series.sindex
    # return result
    return gdf.apply(
        lambda x: _structures_at_intersections(
            x, gdf, sindex, struc_series, struc_sindex, tolerance
        ),
        axis=1,
    )


def no_dangling_node(gdf, datamodel, tolerance):
    """Check if the end-node of a linestring object is not within tolerance of
    a start-node of another linestring object in the same layer.

    Parameters
    ----------
    gdf : ExtendedGeoDataframe
        ExtendedGeoDataFrame, HyDAMO hydroobject layer
    datamodel : HyDAMO
        HyDAMO datamodel class
    tolerance : numeric
        Max tolerance to determine if nodes are connected

    Returns
    -------
    Pandas Series
        Default dtype is bool

    """
    end_nodes_series = gdf["geometry"].apply(lambda x: Point(x.coords[-1]))
    series = gdf["geometry"].apply(lambda x: Point(x.coords[0]))
    sindex = series.sindex

    return end_nodes_series.apply(
        lambda x: _intersects_end_node(x, series, sindex, tolerance)
    )


def structures_at_boundaries(gdf, datamodel, areas, structures, tolerance, distance):
    """
    Check if there are structures near area (typically water-level areas) boundaries.

    Parameters
    ----------
    gdf : ExtendedGeoDataframe
        ExtendedGeoDataFrame, HyDAMO hydroobject layer
    datamodel : HyDAMO
        HyDAMO datamodel class
    areas : str
        HyDAMO datamodel class with areas ("peilgebiedenpraktijk")
    structures : str
        List with structure-types to be expected on the boundary
    tolerance : numeric
        Tolerance to dermine if a structure is on the hydroobject
    distance : numeric
        Max distance between structure and area-boundary

    Returns
    -------
    Pandas Series
        Default dtype is bool

    """

    areas_gdf = getattr(datamodel, areas)
    areas_sindex = areas_gdf.sindex

    struc_series = _layers_from_datamodel(structures, datamodel)
    struc_sindex = struc_series.sindex

    return gdf.apply(
        lambda x: _structures_at_boundaries(
            x, areas_gdf, areas_sindex, struc_series, struc_sindex, tolerance, distance
        ),
        axis=1,
    )


def distant_to_others(gdf, datamodel, distance):
    """
    Check if two objects are sufficient distant to other objects of the same layer.

    Parameters
    ----------
    gdf : ExtendedGeoDataframe
        ExtendedGeoDataFrame, HyDAMO hydroobject layer
    datamodel : HyDAMO
        HyDAMO datamodel class
    distance : numeric
        Max distance to other node

    Returns
    -------
    Pandas Series
        Default dtype is bool

    """

    sindex = gdf.sindex

    return gdf.apply(lambda x: _distant_to_others(x, gdf, sindex, distance), axis=1)


def structures_at_nodes(gdf, datamodel, structures, tolerance):
    """
    Check if structures are on boundary-nodes of lines (hydroobjects)

    Parameters
    ----------
    gdf : ExtendedGeoDataframe
        ExtendedGeoDataFrame, HyDAMO hydroobject layer
    datamodel : HyDAMO
        HyDAMO datamodel class
    structures : str
        List with structure-types to be expected on the boundary
    tolerance : numeric
        Tolerance to determine if a structure is on a node

    Returns
    -------
    Pandas Series
        Default dtype is bool

    """
    struc_series = _layers_from_datamodel(structures, datamodel)
    struc_sindex = struc_series.sindex

    return gdf["geometry"].apply(
        lambda x: _no_struc_on_line(x, struc_series, struc_sindex, tolerance)
    )


def compare_longitudinal(
    gdf,
    datamodel,
    parameter,
    compare_object,
    compare_parameter,
    direction,
    logical_operator,
):
    """
    Check if the value of a parameter in an object is lower/greater than an
    upstream/downstream value of another parameter from another object-layer.


    Parameters
    ----------
    gdf : ExtendedGeoDataframe
        ExtendedGeoDataFrame, HyDAMO hydroobject layer
    datamodel : HyDAMO
        HyDAMO datamodel class
    structures : str
        List with structure-types to be expected on the boundary
    tolerance : numeric
        Tolerance to determine if a structure is on a node

    Returns
    -------
    Pandas Series
        Default dtype is bool

    """
    compare_gdf = getattr(datamodel, compare_object)

    # snap layers to to branches
    compare_gdf.snap_to_branch(datamodel.hydroobject, snap_method="overall")
    gdf.snap_to_branch(datamodel.hydroobject, snap_method="overall")

    return gdf.apply(
        lambda x: _compare_longitudinal(
            x, parameter, compare_gdf, compare_parameter, direction, logical_operator
        ),
        axis=1,
    ).astype(bool)



from pathlib import Path

try:
    from .config import DATA_DIR
except ImportError:
    from config import DATA_DIR

coverage = {"AHN": DATA_DIR.joinpath(r"dtm")}
directory = DATA_DIR.joinpath(r"tasks/test_profielen")
exports_dir = Path(__file__).parent / "exports"
exports_dir.mkdir(exist_ok=True)

# hydamo_validator = validator(output_types=["geopackage", "csv", "geojson"],
#                              coverages=coverage,
#                              log_level="INFO"
#                              )


# datamodel, layer_summary, result_summary = hydamo_validator(
#     directory=directory,
#     raise_error=True
#     )

# datamodel.to_geopackage(exports_dir / "datamodel.gpkg",
#                         use_schema=False)



# %%
from pathlib import Path
from datetime import datetime
from dataclasses import dataclass
import json
from hydamo_validation import validator
import os
import shutil

import hydamo_validation

try:
    from .config import DATA_DIR
except ImportError:
    from config import DATA_DIR

API_VERSION = "0.0"
SCHEMAS_PATH = Path(hydamo_validation.__file__).parent / "schemas"


@dataclass
class User:
    name: str = "test_user"


@dataclass
class Task:
    id: int
    format: str = "geopackage,csv,geojson"


@dataclass
class TaskHelper:
    data_folder: str = DATA_DIR.as_posix()


TaskHelper = TaskHelper()
SCHEMAS_TO_PATH = Path(f"{TaskHelper.data_folder}/schemas")


def execute_task_by_id_and_format(
    self, user: User, task: Task, api_version: str
) -> dict:
    result = {}
    try:
        data = {
            "api_version": api_version,
            "user_name": user.name,
            "task_started": datetime.now().strftime("%Y-%m-%d %H:%M:%S"),
            "task_finished": "",
        }

        directory = f"{TaskHelper.data_folder}/{task.id}"
        covarage = f"{TaskHelper.data_folder}/dtm"

        hydamo_validator = validator(
            output_types=task.format.split(","),
        )

        # start validate
        datamodel, layer_summary, result_summary = hydamo_validator(
            directory=directory, raise_error=False
        )

        result = result_summary.to_dict()

        # end time of validation
        data["task_finished"] = datetime.now().strftime("%Y-%m-%d %H:%M:%S")

        # append api information to the validation result
        results_file = f"{directory}/results/validation_result.json"
        if os.path.exists(results_file):
            with open(results_file, "r+") as file:
                result_data = json.load(file)
                data.update(result_data)

            with open(results_file, "w") as file:
                json.dump(data, file, indent=4)

    except Exception as e:
        result = {"Exception": str(e)}

    with open(f"{directory}/results/result.json", "w") as f:
        f.write(str(result))

    return result


def test_run_productie():
    if SCHEMAS_TO_PATH.exists():
        shutil.rmtree(SCHEMAS_TO_PATH)
    shutil.copytree(SCHEMAS_PATH, f"{TaskHelper.data_folder}/schemas")
    user = User()
    task = Task(1877)

    result = execute_task_by_id_and_format(
        self=None, user=user, task=task, api_version=API_VERSION
    )

    assert result["success"]


# %%



import logging
import geopandas as gpd
import pandas as pd
from pathlib import Path
import json

from hydamo_validation import __version__

OUTPUT_TYPES = ["geopackage", "geojson", "csv"]


class LayersSummary:
    def __init__(self, log_level="DEBUG", date_check=pd.Timestamp.now().isoformat()):
        self.geo_types = {}
        self.date_check = date_check
        self.logger = logging.getLogger(__name__)
        self.logger.setLevel(getattr(logging, log_level))

    def _get_properties(self, gdf):
        properties = {"nen3610id": "str"}
        for i in gdf.columns:
            if i == "syntax_oordeel":
                properties[i] = "bool"
            elif "syntax" in i:
                properties[i] = "str"
            elif "validate" in i:
                properties[i] = "bool"
            elif "general" in i:
                if gdf[i].dtype == float:
                    properties[i] = "float"
                else:
                    properties[i] = "str"
            elif i == "rating":
                properties[i] = "int"
            elif i != "geometry":
                properties[i] = "str"
        return properties

    def set_data(self, gdf, layer, geo_type):
        """
        Set a gdf as a property of results

        Parameters
        ----------
        gdf : GeoDataFrame
            GeoDataFrame that will be property of LayersSummary
        layer : str
            The name of the property the GeoDataFrame will be assigned to. This
            is typically a HyDAMO layer (stuw, hydroobject, ...)
        geo_type : str
            Fiona string representation of geometry type (LineString, Polygon, ...)

        Returns
        -------
        None.

        """

        setattr(self, layer, gdf)
        gdf = getattr(self, layer)
        self.geo_types[layer] = geo_type

    def join_gdf(self, gdf, layer):
        """
        Join a GeoDataFrame to an existing property in layer summary.

        Parameters
        ----------
        gdf : GeoDataFrame
            GeoDataFrame to join to layer
        layer : str
            The property-name this gdf should be linked to. This
            is typically a HyDAMO layer (stuw, hydroobject, ...)

        Returns
        -------
        None.

        """

        if hasattr(self, layer):
            results_gdf = getattr(self, layer)
            drop_cols = [i for i in ["geometry", "nen3610id"] if i in gdf.columns]
            if drop_cols:
                gdf.drop(columns=drop_cols, inplace=True)
            setattr(
                self,
                layer,
                results_gdf.join(gdf),
            )

    def export(self, results_path, output_types=OUTPUT_TYPES):
        """
        Export the content of class to results_path

        Parameters
        ----------
        results_path : str or Path
            Directory where results are to be written to
        output_types : List[str], optional
            The types of output files that will be written. Options are
            ["geojson", "csv", "geopackage"]. By default all will be written
        Returns
        -------
        layers : List(str)
            A list of HyDAMO layers that are successfully written

        """

        gdf_dict = {
            k: v for k, v in self.__dict__.items() if isinstance(v, gpd.GeoDataFrame)
        }
        layers = []
        # make directories for output_types
        results_path = Path(results_path)
        for output_type in ["geojson", "csv"]:
            if output_type in output_types:
                result_dir = results_path.joinpath(output_type)
                result_dir.mkdir(parents=True, exist_ok=True)

        # export results to files
        for object_layer, gdf in gdf_dict.items():
            if "rating" not in gdf.columns:
                gdf["rating"] = 10

            if not gdf.empty:
                schema = {
                    "properties": self._get_properties(gdf),
                    "geometry": self.geo_types[object_layer],
                }

                # add date_check
                gdf["date_check"] = self.date_check
                schema["properties"]["date_check"] = "str"

                for output_type in output_types:
                    # set gdf to WGS84 for export to geojson
                    if output_type == "geojson":
                        file_path = results_path.joinpath(
                            output_type, f"{object_layer}.geojson"
                        )
                        gdf_out = gdf.copy()
                        if gdf_out.crs:
                            gdf_out.to_crs("epsg:4326", inplace=True)
                        gdf_out.to_file(file_path, driver="GeoJSON")

                    # drop geometry for writing to csv
                    elif output_type == "csv":
                        file_path = results_path.joinpath(
                            output_type, f"{object_layer}.csv"
                        )
                        df = gdf.drop("geometry", axis=1)
                        df.to_csv(file_path)

                    # write to geopackage as is
                    elif output_type == "geopackage":
                        file_path = results_path.joinpath("results.gpkg")

                        gdf.to_file(
                            file_path,
                            layer=object_layer,
                            driver="GPKG",
                            engine="pyogrio",
                        )
                layers += [object_layer]
            else:
                self.logger.warn(f"{object_layer} is empty (!)")
        return layers


class ResultSummary:
    def __init__(self, date_check=pd.Timestamp.now().isoformat()):
        """Initialize class."""
        self.success = False
        self.module_version = __version__
        self.date_check = date_check
        self.duration = None
        self.status = "Initialization"
        self.dataset_layers = []
        self.result_layers = []
        self.missing_layers = []
        self.error_layers = []
        self.syntax_result = []
        self.validation_result = []
        self.error = None
        self.errors = None
        self.warnings = None

    def _append_to_list(self, message, property):
        """Append a a message to a list."""
        if getattr(self, property) is None:
            setattr(self, property, [])
        getattr(self, property).append(message)

    def to_json(self, results_path):
        """
        Write result to json

        Parameters
        ----------
        results_path : str or Path
            Directory where results are to be written to

        Returns
        -------
        None.

        """

        result_json = Path(results_path).joinpath("validation_result.json")

        result_dict = {k: v for k, v in self.__dict__.items() if v is not None}
        with open(result_json, "w", encoding="utf-8", newline="\n") as dst:
            json.dump(result_dict, dst, indent=4)

    def to_dict(self):
        """Return class-content as dict."""
        return {k: v for k, v in self.__dict__.items() if v is not None}

    def to_all(self, results_path):
        """
        Write result to json and return class-content as dict

        Parameters
        ----------
        results_path : str or Path
            Directory where results are to be written to

        Returns
        -------
        None.

        """
        self.to_json(results_path)
        return self.to_dict()

    def append_warning(self, message):
        """
        Append a warning

        Parameters
        ----------
        message : str
            Warning message to be appended

        Returns
        -------
        None.

        """

        self._append_to_list(message, "warnings")

    def append_error(self, message):
        """
        Append a error

        Parameters
        ----------
        message : str
            Error message to be appended

        Returns
        -------
        None.

        """
        self._append_to_list(message, "errors")



"""
Test voor __init__.py
"""

from hydamo_validation.__init__ import (
    __author__,
    __copyright__,
    __credits__,
    __version__,
    __license__,
    __maintainer__,
    __email__,
)

author = ["Het Waterschapshuis", "D2HYDRO", "HKV", "HydroConsult"]
copyright = "Copyright 2021, HyDAMO ValidatieTool"
credits = ["D2HYDRO", "HKV", "HydroConsult"]

license = "MIT"
maintainer = "Daniel Tollenaar"
email = "daniel@d2hydro.nl"


def test_author():
    assert __author__ == author


def test_copyright():
    assert __copyright__ == copyright


def test_credits():
    assert __credits__ == credits


def test_license():
    assert __license__ == license


def test_maintainer():
    assert __maintainer__ == maintainer


def test_email():
    assert __email__ == email



from pathlib import Path
import fiona
from hydamo_validation.utils import normalize_fiona_schema, read_geopackage
from hydamo_validation.summaries import ResultSummary
from typing import Union, Dict
import numpy as np


class DataSets:
    """ValidationResult to be dumped as json."""

    def __init__(self, dataset_dir: Union[str, Path]):
        """Initialie datasets."""
        self.path = Path(dataset_dir)
        self.properties: Dict = {}

        self._set_properties()

    def _set_properties(self):
        dataset_files = self.path.glob("*.gpkg")

        for gpkg in dataset_files:
            layers = fiona.listlayers(gpkg)
            layers_dict = {}
            for layer in layers:
                with fiona.open(gpkg, layer=layer) as src:
                    schema = normalize_fiona_schema(src.schema)
                    layers_dict[layer] = schema

            self.properties[gpkg.name] = layers_dict

    def _filter_status(self, gdf, status_object):
        if status_object is not None:
            if "statusobject" in gdf.columns:
                # gdf = gdf.loc[gdf["statusobject"].isin(status_object)]
                gdf = gdf.loc[
                    np.where(
                        gdf["statusobject"].isna()
                        | gdf["statusobject"].isin(status_object)
                    )
                ]
        return gdf

    @property
    def layers(self):
        """Return the layers in the datasets."""
        dataset_layers_dict = {k: list(v.keys()) for k, v in self.properties.items()}

        layers = [
            item for sublist in list(dataset_layers_dict.values()) for item in sublist
        ]

        return layers

    def read_layer(self, layer, result_summary=ResultSummary(), status_object=None):
        """
        Read a layer from the dataset.

        Parameters
        ----------
        layer : str
            Name of the layer (case sensitive!)
        result_summary : ResultSummary
            A hydamo_validation ResultSummary class where a possible exception
            will be appended to.
        status_object : List[str], optional
            A list of statusobject values used as a filter. The default is None.

        Raises
        ------
        e
            General exception while reading the layer from the geopackage.
        KeyError
            Specific exception; the layer is not part of the geopackage.

        Returns
        -------
        gdf : GeoDataFrame
            GeoDataFrame read from datasets (all columns are converted to lower case)
        schema : TYPE
            Fiona schema read from the layer
        """

        if layer in self.layers:
            dataset = {k: v for k, v in self.properties.items() if layer in v.keys()}
            file_path = self.path.joinpath(list(dataset.keys())[0])
            schema = list(dataset.values())[0][layer]
            try:
                gdf = read_geopackage(file_path, layer=layer)
                gdf = self._filter_status(gdf, status_object)
            except Exception as e:
                result_summary.append_warning(
                    (
                        f"Laag {layer} uit bestand {file_path.name} is geen "
                        "GeoPackage die wij kunnen openen. Vervang het bestand en "
                        "probeer opnieuw."
                    )
                )
                raise e

            # we will read all lower case
            schema["properties"] = {
                k.lower(): v for k, v in schema["properties"].items()
            }
            gdf.columns = [i.lower() for i in gdf.columns]
        else:
            raise KeyError(f"'{layer}' not in dataset-layers: '{self.layers}'")

        return gdf, schema



# %%
from hydamo_validation import topologic_functions
from hydamo_validation import validator
from pathlib import Path
from .config import COVERAGE

hydamo_validator = validator(
    output_types=["geopackage", "csv", "geojson"], coverages=COVERAGE, log_level="INFO"
)

directory = Path(r"d:\projecten\D2401.ValidatieModule\01.Issues\GIT-017")

datamodel, layer_summary, result_summary = hydamo_validator(
    directory=directory, raise_error=True
)

gdf = datamodel.hydroobject

# %%
# make a geodataframe from structures list
struc_series = topologic_functions._layers_from_datamodel(
    ["duikersifonhevel"], datamodel
)
# create spatial indices
sindex = gdf.sindex
struc_sindex = struc_series.sindex

for row in gdf.itertuples():
    geometry = row.geometry

    # get a selection of hydroobjects that intersect geometry
    indices = sindex.intersection(geometry.bounds)
    indices = [i for i in indices if i != gdf.index.get_loc(row.Index)]

# %%



from hydamo_validation import topologic_functions
import geopandas as gpd
import pytest

try:
    from .config import DATA_DIR
except ImportError:
    from config import DATA_DIR


@pytest.fixture
def gdf():
    gdf = gpd.read_file(DATA_DIR / "test_layers.gpkg", layer="not_overlapping_line")
    result = topologic_functions.not_overlapping(gdf, datamodel=None, tolerance=0.1)
    gdf.loc[:, ["test_result"]] = result
    gdf.to_file(DATA_DIR / "result_layers.gpkg", layer="not_overlapping_line")
    return gdf.set_index("beschrijving")


@pytest.fixture
def point_gdf():
    gdf = gpd.read_file(DATA_DIR / "test_layers.gpkg", layer="not_overlapping_point")
    result = topologic_functions.not_overlapping(gdf, datamodel=None, tolerance=1)
    gdf.loc[:, ["test_result"]] = result
    gdf.to_file(DATA_DIR / "result_layers.gpkg", layer="not_overlapping_point")
    return gdf.set_index("beschrijving")


def test_line_enkel(gdf):
    assert (gdf.loc["enkel"]["valid"] == gdf.loc["enkel"]["test_result"]).all()


def test_line_aansluitend(gdf):
    assert (
        gdf.loc["aansluitend"]["valid"] == gdf.loc["aansluitend"]["test_result"]
    ).all()


def test_line_kruisend(gdf):
    assert (gdf.loc["kruisend"]["valid"] == gdf.loc["kruisend"]["test_result"]).all()


def test_line_parallel(gdf):
    assert (gdf.loc["parallel"]["valid"] == gdf.loc["parallel"]["test_result"]).all()


def test_line_aantakkend(gdf):
    assert (
        gdf.loc["aantakkend"]["valid"] == gdf.loc["aantakkend"]["test_result"]
    ).all()


def test_line_volledig_overlappend_1(gdf):
    assert (
        gdf.loc["volledig overlappend 1"]["valid"]
        == gdf.loc["volledig overlappend 1"]["test_result"]
    ).all()


def test_line_volledig_overlappend_2(gdf):
    assert (
        gdf.loc["volledig overlappend 2"]["valid"]
        == gdf.loc["volledig overlappend 2"]["test_result"]
    ).all()


def test_line_deels_overlappend(gdf):
    assert (
        gdf.loc["deels overlappend"]["valid"]
        == gdf.loc["deels overlappend"]["test_result"]
    ).all()


def test_line_buiten_tolerantie(gdf):
    assert (
        gdf.loc["buiten tolerantie"]["valid"]
        == gdf.loc["buiten tolerantie"]["test_result"]
    ).all()


def test_line_binnen_tolerantie(gdf):
    assert (
        gdf.loc["binnen tolerantie"]["valid"]
        == gdf.loc["binnen tolerantie"]["test_result"]
    ).all()


def test_line_complete_gdf(gdf):
    assert (gdf["valid"] == gdf["test_result"]).all()


def test_point_enkel(point_gdf):
    assert (
        point_gdf.loc["enkel"]["valid"] == point_gdf.loc["enkel"]["test_result"]
    ).all()


def test_point_buiten_tolerantie(point_gdf):
    assert (
        point_gdf.loc["buiten tolerantie"]["valid"]
        == point_gdf.loc["buiten tolerantie"]["test_result"]
    ).all()


def test_point_binnen_tolerantie(point_gdf):
    assert (
        point_gdf.loc["binnen tolerantie"]["valid"]
        == point_gdf.loc["binnen tolerantie"]["test_result"]
    ).all()


def test_point_complete_gdf(point_gdf):
    assert (point_gdf["valid"] == point_gdf["test_result"]).all()



import fiona
import time
import logging
import geopandas as gpd
import pandas as pd
from inspect import getmembers, isfunction


def normalize_fiona_schema(schema):
    schema["properties"] = {
        k: fiona.schema.normalize_field_type(v) for k, v in schema["properties"].items()
    }
    return schema


def schema_properties_to_dtypes(properties):
    properties = {
        k: fiona.schema.normalize_field_type(v) for k, v in properties.items()
    }
    return properties


def dataset_layers(dataset_properties):
    dataset_layers_dict = {k: list(v.keys()) for k, v in dataset_properties.items()}

    layers = [
        item for sublist in list(dataset_layers_dict.values()) for item in sublist
    ]

    return layers


def get_functions(module):
    return [i[0] for i in getmembers(module, isfunction) if i[0][0] != "_"]


def read_geopackage(file_path, layer):
    """Read file as GeoDataFrame."""
    gdf = gpd.read_file(file_path, layer=layer, engine="pyogrio", use_fid_as_index=True)

    if type(gdf) == pd.DataFrame:
        gdf = gpd.GeoDataFrame(gdf, geometry=gpd.GeoSeries())

    # fix integers to nullable type
    with fiona.open(file_path, "r", layer=layer) as src:
        dtypes = normalize_fiona_schema(src.schema)["properties"]

    for k, v in dtypes.items():
        if v == "int64":
            gdf[k] = gdf[k].astype(pd.Int64Dtype())

    return gdf


class Timer(object):
    """Record function efficiency."""

    def __init__(self, logger=logging):
        self.start = time.time()
        self.milestone = self.start
        self.logger = logger

    def report(self, message=""):
        """Set milestone and report."""
        delta_time = time.time() - self.milestone
        self.logger.debug(f"{message} in {delta_time:.3f} sec")
        self.milestone = time.time()
        return delta_time

    def reset(self, message=None):
        """Report task-efficiency and reset."""
        if message:
            self.logger.debug(f"{message} in {(time.time() - self.start):.3f} sec")
        self.start = time.time()
        self.milestone = self.start



"""HyDAMO datamodel for ValidatieTool."""

import json
import logging
import re
import warnings
from pathlib import Path
from typing import Any, Literal

import fiona
import geopandas as gpd
from geopandas import GeoSeries
import numpy as np
from shapely.geometry import LineString, MultiLineString, MultiPolygon, Point, Polygon

from hydamo_validation import geometry
from hydamo_validation.styles import add_styles_to_geopackage

warnings.filterwarnings("ignore", category=UserWarning)
warnings.filterwarnings("ignore", category=RuntimeWarning)
warnings.filterwarnings("ignore", category=FutureWarning)

FIELD_TYPES_MAP_REV = fiona.schema.FIELD_TYPES_MAP_REV
FIELD_TYPES_MAP = fiona.schema.FIELD_TYPES_MAP
MODEL_CRS = "epsg:28992"
SCHEMAS_DIR = Path(__file__).parent.joinpath("schemas", "hydamo")

GEOTYPE_MAPPING = {
    "LineString": LineString,
    "MultiLineString": MultiLineString,
    "Point": Point,
    "PointZ": Point,
    "Polygon": Polygon,
    "MultiPolygon": MultiPolygon,
}

DTYPE_MAPPING = {
    "string": "str",
    "integer": "int64",
    "date-time": "datetime",
    "number": "float",
}

default_properties: dict[str, Any] = {
    "id": None,
    "dtype": "str",
    "required": False,
    "unique": False,
}


def map_definition(definition: dict[str, Any]) -> list[dict[str, Any]]:
    """


    Parameters
    ----------
    definition : dict[str, Any]
        HyDAMO definition as specified in the HyDAMO JSON specification.

    Returns
    -------
    list[dict[str, Any]]
        Validation schema for the HyDAMO class.

    """

    # start with an empty list to populate
    result = []

    for k, v in definition.items():
        # convert geometry if shape
        if k == "shape":
            properties: dict[str, Any] = {"id": "geometry"}
            dtype = v["type"]
            if not isinstance(dtype, list):
                dtype = [dtype]
            properties["dtype"] = dtype
            properties["required"] = True
            result.append(properties)

        # set properties if not shape
        else:
            properties = default_properties.copy()
            properties["id"] = k
            if "type" in v.keys():
                properties["dtype"] = DTYPE_MAPPING[v["type"]]
                if "format" in v.keys():
                    properties["dtype"] = DTYPE_MAPPING[v["format"]]

            # set required
            if "minItems" in v.keys():
                if v["minItems"] == 1:
                    properties["required"] = True

            # set unique
            if "uniqueItems" in v.keys():
                properties["unique"] = v["uniqueItems"]

            # set domain
            if "enum" in v.keys():
                properties["domain"] = [{"value": i} for i in v["enum"]]
            result.append(properties)
    return result


class ExtendedGeoDataFrame(gpd.GeoDataFrame):  # type: ignore
    """A GeoPandas GeoDataFrame with extended properties and methods."""

    # ignores subclassing Any: https://github.com/geopandas/geopandas/discussions/2750

    _metadata = [
        "required_columns",
        "geotype",
        "layer_name",
    ] + gpd.GeoDataFrame._metadata

    def __init__(
        self,
        validation_schema: list[dict[str, Any]],
        geotype: (
            list[
                Literal[
                    "LineString",
                    "MultiLineString",
                    "Point",
                    "PointZ",
                    "Polygon",
                    "MultiPolygon",
                ]
            ]
            | None
        ),
        layer_name: str = "",
        required_columns: list[str] = [],
        logger=logging,
        *args,
        **kwargs,
    ):
        # Check type
        required_columns = [i.lower() for i in required_columns]

        # Add required columns to column list
        kwargs["columns"] = required_columns
        kwargs["geometry"] = GeoSeries()

        super().__init__(*args, **kwargs)

        self.validation_schema = validation_schema
        self.required_columns = required_columns
        self.layer_name = layer_name
        self.geotype = geotype

        if "geometry" not in self.required_columns:
            self.required_columns += ["geometry"]
        self.crs = MODEL_CRS

    def _check_columns(self, gdf):
        """Check presence of columns in GeoDataFrame"""
        present_columns = gdf.columns.tolist()
        for column in self.required_columns:
            if column not in present_columns:
                raise KeyError(
                    'Column "{}" not found. Got {}, Expected at least {}'.format(
                        column,
                        ", ".join(present_columns),
                        ", ".join(self.required_columns),
                    )
                )

    def _check_geotype(self):
        """Check geometry type"""
        if self.geotype:
            if not all(
                any(isinstance(geo, GEOTYPE_MAPPING[i]) for i in self.geotype)
                for geo in self.geometry
            ):
                raise TypeError(
                    'Geometry-type "{}" required in layer "{}". The input feature-file has geometry type(s) {}.'.format(
                        re.findall("([A-Z].*)'", repr(self.geotype))[0],
                        self.layer_name,
                        self.geometry.type.unique().tolist(),
                    )
                )

    def _get_schema(self):
        """Return fiona schema dict from validation_schema."""
        properties = {
            i["id"]: i["dtype"] for i in self.validation_schema if i["id"] != "geometry"
        }
        # properties = {k: (v if v != "datetime" else "str") for k, v in properties}
        geometry = next(
            (i["dtype"] for i in self.validation_schema if i["id"] == "geometry"),
            None,
        )
        return {"properties": properties, "geometry": geometry}

    def set_data(
        self,
        gdf,
        layer="",
        index_col=None,
        check_columns=True,
        check_geotype=True,
        extra_attributes={},
    ):
        """


        Parameters
        ----------
        gdf : GeoDataFrame
            GeoDataFrame with a HyDAMO object-layer
        index_col : str, optional
            Column to be used as index. The default is None.
        check_columns : bool, optional
            Check if all required columns are present in the GeoDataFrame.
            The default is True.
        check_geotype : bool, optional
            Check if the geometry is of the required type. The default is True.

        Returns
        -------
        None.

        """

        if not self.empty:
            self.delete_all()

        # reproject to crs if necessary
        if (gdf.crs is not None) and ("geometry" in self.required_columns):
            if f"epsg:{gdf.crs.to_epsg()}" == MODEL_CRS:
                gdf.set_crs(MODEL_CRS, inplace=True, allow_override=True)
            else:
                gdf.to_crs(MODEL_CRS, inplace=True)

        # Check columns
        gdf.columns = [i.lower() for i in gdf.columns]
        if check_columns:
            self._check_columns(gdf)

        # Copy content
        for col, values in gdf.items():
            self[col] = values.to_numpy()

        if index_col is None:
            self.index = gdf.index
            self.index.name = gdf.index.name

        else:
            self.index = gdf[index_col]
            self.index.name = index_col

        # Check geometry types
        if check_geotype:
            self._check_geotype()

        # Set extra attribute-values
        for k, v in extra_attributes.items():
            if k not in self.columns:
                self[k] = v

    def delete_all(self):
        """Empty the dataframe"""
        if not self.empty:
            self.iloc[:, 0] = np.nan
            self.dropna(inplace=True)

    def snap_to_branch(self, branches, snap_method, maxdist=5):
        """


        Parameters
        ----------
        branches : GeoDataFrame
            GeoDataFrame with branches
        snap_method : str
            Options for snapping
        maxdist : float, optional
            The maximal distance for snapping. The default is 5.

        Returns
        -------
        None.

        """

        """Snap the geometries to the branch."""
        geometry.find_nearest_branch(
            branches=branches, geometries=self, method=snap_method, maxdist=maxdist
        )


class HyDAMO:
    """Definition of the HyDAMO datamodel."""

    def __init__(
        self,
        version: str = "2.2",
        schemas_path: Path = SCHEMAS_DIR,
        ignored_layers: list[str] = [
            "afvoeraanvoergebied",
            "imwa_geoobject",
            "leggerwatersysteem",
            "leggerwaterveiligheid",
            "waterbeheergebied",
        ],
    ):
        self.version = version
        self.schema_json = schemas_path.joinpath(f"HyDAMO_{version}.json")
        self.layers: list[str] = []
        self.ignored_layers = ignored_layers

        self.init_datamodel()

    @property
    def data_layers(self):
        return [layer for layer in self.layers if not getattr(self, layer).empty]

    def init_datamodel(self) -> None:
        """Initialize DataModel from self.schemas_path."""
        self.validation_schemas: dict[str, Any] = {}

        # read schema as dict
        with open(self.schema_json) as src:
            schema = json.load(src)
            hydamo_layers = [
                Path(i["$ref"]).name for i in schema["properties"]["HyDAMO"]["anyOf"]
            ]
            self.layers = [i for i in hydamo_layers if i not in self.ignored_layers]

        for hydamo_layer in self.layers:
            definition = schema["definitions"][hydamo_layer]["properties"]
            layer_schema = map_definition(definition)
            self.validation_schemas[hydamo_layer] = layer_schema

            # add layer to data_model
            geotype = next(
                (i["dtype"] for i in layer_schema if i["id"] == "geometry"), None
            )

            required_columns = [
                i["id"]
                for i in [i for i in layer_schema if "required" in i.keys()]
                if i["required"]
            ]

            setattr(
                self,
                hydamo_layer,
                ExtendedGeoDataFrame(
                    validation_schema=layer_schema,
                    layer_name=hydamo_layer,
                    geotype=geotype,
                    required_columns=required_columns,
                ),
            )

    def get(self, layer: str, global_id: str):
        """
        Get a DataFrame row (feature) providing a layer an global_id.

        Parameters
        ----------
        layer : str
            DESCRIPTION.
        global_id : str
            DESCRIPTION.

        Returns
        -------
        TYPE
            DESCRIPTION.

        """
        return getattr(self, layer).set_index("globalid").loc[global_id]

    def set_data(
        self,
        gdf,
        layer,
        index_col=None,
        check_columns=True,
        check_geotype=True,
        extra_attributes={},
    ):
        """


        Parameters
        ----------
        gdf : GeoDataFrame
            GeoDataFrame with a HyDAMO object-layer
        layer : TYPE
            HyDAMO layer to be set
        index_col : str, optional
            Column to be used as index. The default is None.
        check_columns : bool, optional
            Check if all required columns are present in the GeoDataFrame.
            The default is True.
        check_geotype : bool, optional
            Check if the geometry is of the required type. The default is True.

        Returns
        -------
        None.

        """

        getattr(self, layer).set_data(
            gdf,
            index_col=index_col,
            check_columns=check_columns,
            check_geotype=check_geotype,
            extra_attributes={},
        )

    def to_geopackage(self, file_path, use_schema=True):
        """

        Parameters
        ----------
        file_path : path-string
            Path-string where the file should be written to
        use_schema : bool, optional
            Use the schema to specify column-properties The default is True.

        Returns
        -------
        None.

        """
        file_path = Path(file_path)
        for layer in self.layers:
            gdf = getattr(self, layer).copy()
            if not gdf.empty:
                if use_schema:
                    # match fiona layer schema keys with gdf.columns
                    schema = getattr(self, layer)._get_schema()
                    schema_cols = list(schema["properties"].keys()) + ["geometry"]
                    drop_cols = [i for i in gdf.columns if i not in schema_cols]
                    gdf.drop(columns=drop_cols, inplace=True)

                    schema["properties"] = {
                        k: v
                        for k, v in schema["properties"].items()
                        if k in gdf.columns
                    }

                    # write gdf to geopackage, including schema
                    if gdf.index.name in gdf.columns:
                        gdf.reset_index(drop=True, inplace=True)
                    gdf.to_file(
                        file_path,
                        layer=layer,
                        driver="GPKG",
                        engine="pyogrio",
                    )
                else:
                    # write gdf to geopackage as is
                    if gdf.index.name in gdf.columns:
                        gdf = gdf.reset_index(drop=True).copy()
                    gdf.to_file(file_path, layer=layer, driver="GPKG", engine="pyogrio")
        if file_path.is_file():
            add_styles_to_geopackage(file_path)

    @classmethod
    def from_geopackage(
        cls, file_path, version="2.2", check_columns=True, check_geotype=True
    ):
        """
        Initialize HyDAMO class from GeoPackage

        Parameters
        ----------
        file_path : path-string
            Path-string to the hydamo GeoPackage
        check_columns : bool, optional
            Check if all required columns are present in the GeoDataFrame.
            The default is True.
        check_geotype : bool, optional
            Check if the geometry is of the required type. The default is True.

        Returns
        -------
        hydamo : HyDAMO
            HyDAMO object initialized with content of GeoPackage

        """
        hydamo = cls(version=version)
        for layer in fiona.listlayers(file_path):
            if layer in hydamo.layers:
                hydamo_layer = getattr(hydamo, layer)
                hydamo_layer.set_data(
                    gpd.read_file(file_path, layer=layer),
                    check_columns=check_columns,
                    check_geotype=check_geotype,
                )
        return hydamo



import re
import sqlite3
from datetime import datetime
from pathlib import Path

import fiona

STYLES_DIR = Path(__file__).parent.joinpath("data", "styles")

CREATE_TABLE_SQL = """
CREATE TABLE "layer_styles" (
    "id" INTEGER PRIMARY KEY AUTOINCREMENT NOT NULL,
    "f_table_catalog" TEXT(256),
    "f_table_schema" TEXT(256),
    "f_table_name" TEXT(256),
    "f_geometry_column" TEXT(256),
    "styleName" TEXT(30),
    "styleQML" TEXT,
    "styleSLD" TEXT,
    "useAsDefault" BOOLEAN,
    "description" TEXT,
    "owner" TEXT(30),
    "ui" TEXT(30),
    "update_time" DATETIME DEFAULT (strftime('%Y-%m-%dT%H:%M:%fZ','now'))
);
"""
DROP_TABLE_SQL = """DROP TABLE IF EXISTS "layer_styles";"""

INSERT_ROW_SQL = """
INSERT INTO "main"."layer_styles" (
    "f_table_catalog",
    "f_table_schema",
    "f_table_name",
    "f_geometry_column",
    "styleName",
    "styleQML",
    "styleSLD",
    "useAsDefault",
    "description",
    "owner",
    "ui",
    "update_time"
)
VALUES (
    '',
    '',
    '{layer}',
    'geom',
    '{layer}',
    '{style_qml}',
    '{style_sld}',
    '1',
    '{description}',
    '',
    '',
    '{update_date_time}'
);
"""


def read_style(style_path: Path) -> str:
    """
    To make style-text sql-compatible, we need to replace single ' to ''.
    Example 'http://mrcc.com/qgis.dtd -> ''http://mrcc.com/qgis.dtd''

    Parameters
    ----------
    style_path : Path
        Path to sld-file

    Returns
    -------
    str
        style-string for SQL

    """
    style_txt = style_path.read_text()

    pattern = r"'(.*?)'"
    style_txt = re.sub(pattern, lambda m: f"''{m.group(1)}''", style_txt)

    return style_txt


def add_styles_to_geopackage(gpkg_path: Path):
    """
    Add styles to a HyDAMO GeoPackage

    Parameters
    ----------
    gpkg_path : Path
        Path to HyDAMO GeoPackage

    Returns
    -------
    None.

    """

    with sqlite3.connect(gpkg_path) as conn:
        # create table
        conn.execute(DROP_TABLE_SQL)
        conn.execute(CREATE_TABLE_SQL)

        # add style per layer
        for layer in fiona.listlayers(gpkg_path):
            style_qml = STYLES_DIR / f"{layer}.qml"
            style_sld = STYLES_DIR / f"{layer}.sld"

            # check if style exists
            if style_qml.exists() and style_sld.exists():
                description = f"HyDAMO style for layer: {layer}"
                update_date_time = f"{datetime.now().isoformat()}Z"

                # push to GeoPackage
                conn.execute(
                    INSERT_ROW_SQL.format(
                        layer=layer,
                        style_qml=read_style(style_qml),
                        style_sld=read_style(style_sld),
                        description=description,
                        update_date_time=update_date_time,
                    )
                )



"""Syntax validation."""

import numpy as np
import pandas as pd
from fiona.schema import normalize_field_type


# %% validation functions applied on datasets
def datamodel_layers(dm_layers, ds_layers):
    """Find non-datamodel layers in datasets."""
    layers = [i for i in ds_layers if dm_layers.count(i.lower()) == 1]
    return layers


def missing_layers(dm_layers, ds_layers):
    """Find missing datamodel-layers in datasets."""
    layers = [i.lower() for i in ds_layers]
    layers = [i for i in dm_layers if i not in layers]
    return layers


def duplicate_layers(layers):
    """Find duplicate layers in datasets."""
    error_layers = list(set([item for item in layers if layers.count(item) > 1]))
    error = {}
    critical = False

    # clean layers and report error
    if error_layers:
        error_layers_str = ",".join(error_layers)
        layers = [i for i in layers if i not in error_layers]
        error_layers_str = ",".join(error_layers)
        error = {
            "type": "critical",
            "message": f"datasets bevatten lagen vaker dan 1x die worden verwijderd: {error_layers_str}",
        }

    return layers, error, critical


# %% validation functions applied on dataframes
def non_unique_values(series):
    """Return non-unique values in series."""
    return series.duplicated(keep=False)


def missing_values(series, fill_values=[]):
    """Return non-unique values in series."""
    return series.isnull() | series.isin(fill_values)


def fill_values(series, dtype, fill_values=[-999]):
    """Return filled values in series."""
    return series.isin(np.array(fill_values).astype(dtype))


def non_domain_values(series, dtype, domain_values):
    """Return non-domain values in series."""
    bool_series = ~series.isin(np.array(domain_values).astype(dtype))
    bool_series.loc[series.isna()] = False  # for not na values
    return bool_series


def _convertable_dtype(x, dtype):
    try:
        if dtype == "datetime":
            return False
        elif dtype in ["int", "int64", "float"]:
            if x is None:
                return False
            result = pd.to_numeric(x)
            if dtype in ["int", "int64"]:
                return "." not in str(result)
        return True
    except ValueError:
        return False


def convertable_dtypes(series, dtype):
    """Find convertable dtypes."""
    return series.apply(_convertable_dtype, args=(dtype,))


def _get_constant_series(gdf, value=True, dtype=bool):
    return pd.Series([value] * len(gdf), index=gdf.index, dtype=dtype)


# %% to execute it all


# all fields syntax validation
def fields_syntax(gdf, schema, validation_schema, keep_columns=[]):
    """Validate fields in a gdf to a validation-schema."""
    # initialize results
    if "geometry" in gdf.columns:
        columns_tuple = tuple(keep_columns + ["geometry"])
    else:
        columns_tuple = tuple(keep_columns)
    validation_gdf = gdf.loc[:, columns_tuple].copy()
    result_gdf = gdf.copy()
    valid_summary = _get_constant_series(result_gdf)

    # check iteratively if column fails a valiation rule
    for col in [i for i in validation_schema if i["id"] != "geometry"]:
        valid_series = _get_constant_series(result_gdf)

        replace_series = _get_constant_series(result_gdf, value="", dtype=str)

        col_exists = col["id"] in result_gdf.columns
        result_col = f"syntax_{col['id']}"
        validation_gdf[result_col] = [[] for _ in range(len(validation_gdf))]

        if (not col_exists) and ("required" in col.keys()):
            # required columns should not be missing!
            if col["required"]:
                validation_gdf.loc[:, result_col].apply(lambda x: x.append(4))
                valid_series.loc[:] = False
                if col["dtype"] == "float":
                    result_gdf.loc[:, col["id"]] = np.NaN
                else:
                    result_gdf.loc[:, col["id"]] = pd.NA

        # first check if a column as the correct dtype
        else:
            dtype_fixed = True
            dtype = schema["properties"][col["id"]]
            if dtype != col["dtype"]:
                # try to convert it into the correct data-type
                convertable_rows = convertable_dtypes(
                    result_gdf.loc[:, col["id"]], col["dtype"]
                )

                # replace un-convertable rows to pd.NA so they can be converted
                replace_series[~convertable_rows] = result_gdf.loc[
                    ~convertable_rows, col["id"]
                ].apply(lambda x: f"{x} -> NULL ")
                if col["dtype"] == "float":
                    result_gdf.loc[~convertable_rows, col["id"]] = np.NaN
                else:
                    result_gdf.loc[~convertable_rows, col["id"]] = pd.NA

                if col["dtype"] in ["int", "int64"]:
                    result_gdf.loc[:, col["id"]] = pd.to_numeric(
                        result_gdf[col["id"]]
                    ).astype(pd.Int64Dtype())
                elif col["dtype"] == "datetime":
                    result_gdf.loc[:, col["id"]] = pd.to_numeric(
                        result_gdf[col["id"]]
                    ).astype("datetime64[ns]")
                else:
                    result_gdf.loc[:, col["id"]] = result_gdf[col["id"]].astype(
                        col["dtype"]
                    )
                validation_gdf.loc[convertable_rows, result_col].apply(
                    lambda x: x.append(2)
                )
                # mark all un-convertable rows as un-convertable
                validation_gdf.loc[~convertable_rows, result_col].apply(
                    lambda x: x.append(1)
                )
                valid_series.loc[~convertable_rows] = False
            else:
                convertable_rows = _get_constant_series(result_gdf)
                if normalize_field_type(dtype) == "float":
                    result_gdf[col["id"]] = result_gdf[col["id"]].astype(float)
                elif normalize_field_type(dtype) == "int64":
                    result_gdf[col["id"]] = result_gdf[col["id"]].astype(
                        pd.Int64Dtype()
                    )

            if dtype_fixed:
                # only unique values if specified
                if "unique" in col.keys():
                    if col["unique"] == True:
                        bool_series = non_unique_values(result_gdf[col["id"]])
                        validation_gdf.loc[bool_series, result_col].apply(
                            lambda x: x.append(3)
                        )
                        valid_series.loc[bool_series] = False

                # non missing if required
                if "required" in col.keys():
                    if col["required"] == True:
                        if col_exists:
                            nan_values = []
                            if "domain" in col.keys():
                                nan_values = [
                                    i["value"]
                                    for i in col["domain"]
                                    if i["value"] in [98, 99]
                                ]
                            bool_series = missing_values(
                                result_gdf[col["id"]], fill_values=nan_values
                            )
                            validation_gdf.loc[bool_series, result_col].apply(
                                lambda x: x.append(4)
                            )
                            valid_series.loc[bool_series] = False
                        else:
                            validation_gdf.loc[:, result_col].apply(
                                lambda x: x.append(4)
                            )
                            valid_series.loc[:] = False

                # return filled values
                if col["dtype"] != "datetime":
                    bool_series = fill_values(result_gdf[col["id"]], col["dtype"])
                    validation_gdf.loc[bool_series, result_col].apply(
                        lambda x: x.append(5)
                    )
                    valid_series.loc[bool_series] = False

                # return non-domain values
                if "domain" in col.keys():
                    domain_values = [i["value"] for i in col["domain"]]
                    bool_series = non_domain_values(
                        result_gdf[col["id"]], col["dtype"], domain_values
                    )
                    validation_gdf.loc[bool_series, result_col].apply(
                        lambda x: x.append(6)
                    )
                    valid_series.loc[bool_series] = False

        valid_summary.loc[~valid_series] = False
        # %%
        # set all invalid data from original gdf to Null
        if col_exists:
            # valid_series[~convertable_rows] = True
            replace_series.loc[~valid_series] = result_gdf.loc[
                ~valid_series, col["id"]
            ].apply(lambda x: f"{x} -> NULL ")
            if col["dtype"] == "float":
                result_gdf.loc[~valid_series, col["id"]] = np.NaN
            else:
                result_gdf.loc[~valid_series, col["id"]] = pd.NA

        # map list to string
        validation_gdf.loc[:, result_col] = validation_gdf[result_col].apply(
            lambda x: f'{",".join(map(str, x))}'
        )
        validation_gdf.loc[:, result_col] = validation_gdf[result_col].apply(
            lambda x: f"(fouten: {x})" if len(x) > 0 else x
        )
        validation_gdf.loc[:, result_col] = replace_series.str.cat(
            validation_gdf[result_col]
        )

    # check for invalid geometries and delete these geometries
    geotype = next(
        (i["dtype"] for i in validation_schema if i["id"] == "geometry"), None
    )
    if geotype:
        result_col = "syntax_geometry"
        validation_gdf[result_col] = ""

        bool_series = ~(
            ~result_gdf["geometry"].is_valid | result_gdf["geometry"].type.isin(geotype)
        )
        validation_gdf.loc[bool_series, result_col] = result_gdf[
            bool_series
        ].geometry.apply(lambda x: f"{x.type} -> NULL (7)")
        result_gdf = result_gdf.loc[~bool_series]
        valid_series.loc[bool_series] = False
        valid_summary.loc[bool_series] = False

    # report if any validation error occured
    validation_gdf.loc[:, "syntax_oordeel"] = valid_summary.astype(bool)

    return result_gdf, validation_gdf



"""Logic functions to be used in eval-method."""

import pandas as pd


def _overlapping_period(row, df, start_date, end_date):
    _df = df[df.index != row.name]
    return ~(
        (row[start_date] <= _df[end_date]) & (row[end_date] >= _df[end_date])
    ).any()


def _check_attributes(gdf, attributes):
    for i in attributes:
        if type(i) == str:
            if not i in gdf.columns:
                raise KeyError(
                    rf"'{i}' not in columns: {gdf.columns.to_list()}. Rule cannot be executed"
                )


def LE(gdf, left, right, dtype=bool):
    """
    Evaluate if left is less or equal to/than right

    Parameters
    ----------
    gdf : GeoDataFrame
        Input GeoDataFrame
    left : str, numeric
        Left column or value in expression
    right : TYPE
        Right column or value in expression
    dtype : dtype, optional
        dtype assigned to result Series
        The default is bool.

    Returns
    -------
    result : Series
        Pandas Series (default dtype = bool)

    """
    _check_attributes(gdf, [left, right])
    expression = f"{left} <= {right}".lower()
    return gdf.eval(expression).astype(dtype)


def LT(gdf, left, right, dtype=bool):
    """
    Evaluate if left is less than right

    Parameters
    ----------
    gdf : GeoDataFrame
        Input GeoDataFrame
    left : str, numeric
        Left column or value in expression
    right : TYPE
        Right column or value in expression
    dtype : dtype, optional
        dtype assigned to result Series
        The default is bool.

    Returns
    -------
    result : Series
        Pandas Series (default dtype = bool)

    """
    _check_attributes(gdf, [left, right])
    expression = f"{left} < {right}".lower()
    return gdf.eval(expression).astype(dtype)


def GT(gdf, left, right, dtype=bool):
    """
    Evaluate if left is greater than right

    Parameters
    ----------
    gdf : GeoDataFrame
        Input GeoDataFrame
    left : str, numeric
        Left column or value in expression
    right : TYPE
        Right column or value in expression
    dtype : dtype, optional
        dtype assigned to result Series
        The default is bool.

    Returns
    -------
    result : Series
        Pandas Series (default dtype = bool)

    """
    _check_attributes(gdf, [left, right])
    expression = f"{left} > {right}".lower()
    return gdf.eval(expression).astype(dtype)


def GE(gdf, left, right, dtype=bool):
    """Evaluate if left is greater or equal to/than right

    Parameters
    ----------
    gdf : GeoDataFrame
        Input GeoDataFrame
    left : str, numeric
        Left column or value in expression
    right : TYPE
        Right column or value in expression
    dtype : dtype, optional
        dtype assigned to result Series
        The default is bool.

    Returns
    -------
    result : Series
        Pandas Series (default dtype = bool)

    """
    _check_attributes(gdf, [left, right])
    expression = f"{left} >= {right}".lower()
    return gdf.eval(expression).astype(dtype)


def EQ(gdf, left, right, dtype=bool):
    """Evalate if left an right expression are equal

    Parameters
    ----------
    gdf : GeoDataFrame
        Input GeoDataFrame
    left : str, numeric
        Left column or value in expression
    right : TYPE
        Right column or value in expression
    dtype : dtype, optional
        dtype assigned to result Series
        The default is bool.

    Returns
    -------
    result : Series
        Pandas Series (default dtype = bool)

    """
    _check_attributes(gdf, [left, right])
    expression = f"{left} == {right}".lower()
    return gdf.eval(expression).astype(dtype)


def BE(gdf, parameter, min, max, inclusive=False):
    """Evaluate if parameter-value is between min/max inclusive (true/false)

    Parameters
    ----------
    gdf : GeoDataFrame
        Input GeoDataFrame
    parameter: str
        Input column with numeric values
    min : numeric
        Lower limit of function
    max : numeric
        Upper limit of function
    inclusive : bool, optional
        To include min and max
        The default is False.

    Returns
    -------
    result : Series
        Pandas Series (default dtype = bool)

    """
    _check_attributes(gdf, [parameter, min, max])
    if inclusive:
        series = GE(gdf, parameter, min, dtype=bool) & LE(
            gdf, parameter, max, dtype=bool
        )
    else:
        series = GT(gdf, parameter, min, dtype=bool) & LT(
            gdf, parameter, max, dtype=bool
        )
    return series


def ISIN(gdf, parameter, array):
    """Evaluate if values in parameter are in array

    Parameters
    ----------
    gdf : GeoDataFrame
        Input GeoDataFrame
    parameter: str
        Input column with numeric values
    array : list
        list of possible values that return True

    Returns
    -------
    result : Series
        Pandas Series (default dtype = bool)

    """
    _check_attributes(gdf, [parameter])
    return gdf[parameter].isin(array)


def NOTIN(gdf, parameter, array):
    """Evaluate if values in parameter are not in array

    Parameters
    ----------
    gdf : GeoDataFrame
        Input GeoDataFrame
    parameter: str
        Input column with numeric values
    array : list
        list of possible values that return False

    Returns
    -------
    result : Series
        Pandas Series (default dtype = bool)

    """
    _check_attributes(gdf, [parameter])
    return ~ISIN(gdf, parameter, array)


def NOTNA(gdf, parameter):
    """Evaluate if values in parameter ar not NaN or None

    Parameters
    ----------
    gdf : GeoDataFrame
        Input GeoDataFrame
    parameter: str
        Input column with numeric values

    Returns
    -------
    result : Series
        Pandas Series (default dtype = bool)

    """
    _check_attributes(gdf, [parameter])
    return gdf[parameter].notna()


def join_object_exists(gdf, join_gdf, join_object):
    """Evaluate if defined related_object id exists in globalid parameter of
    related object-table.

    Parameters
    ----------
    gdf : GeoDataFrame
        Input GeoDataFrame
    related_gdf : GeoDataFrame
        Input GeoDataFrame with related objects
    object: str
        HyDAMO object name of related object-layer

    Returns
    -------
    result : Series
        Pandas Series (default dtype = bool)

    """
    _check_attributes(join_gdf, ["globalid"])
    _check_attributes(gdf, [f"{join_object}id"])
    return gdf[f"{join_object}id"].isin(join_gdf["globalid"])


def consistent_period(
    gdf,
    max_gap=1,
    groupers=["pompid", "regelmiddelid"],
    priority="prioriteit",
    start_date="beginperiode",
    date_format="%d%m",
    end_date="eindperiode",
):
    """Check if a periodic-based table is time-consistent

    Parameters
    ----------
    gdf : GeoDataFrame
        Input GeoDataFrame
    max_gap: int
        max gap in days between too adjacent periods

    Returns
    -------
    result : Series
        Pandas Series (default dtype = bool)

    """

    # create an empty result
    _gdf = gdf.copy()
    result = pd.Series(index=_gdf.index)

    # convert start_parameter and end_parameter to datetime
    _gdf[start_date] = pd.to_datetime(_gdf[start_date], format=date_format)
    _gdf[end_date] = pd.to_datetime(_gdf[end_date], format=date_format)

    index_select = _gdf[start_date] > _gdf[end_date]
    _gdf.loc[index_select, end_date] = _gdf[index_select][
        end_date
    ] + pd.offsets.DateOffset(years=1)

    for group in groupers:
        grouper = _gdf.groupby(by=[group, "prioriteit"])

        for _, df in _gdf.groupby(by=["pompid", "prioriteit"]):
            df.sort_values(by=start_date, inplace=True)

            # check for overlap
            bool_series = df.apply(
                (lambda x: _overlapping_period(x, df, start_date, end_date)), axis=1
            )

            # check for gaps
            gaps_series = df[start_date] - df.shift(1)[end_date]
            gaps_series.iloc[0] = pd.Timedelta(days=0)  # due to shift we have NaT here

            # add to result
            bool_series = (gaps_series <= pd.Timedelta(days=int(max_gap))) & bool_series
            bool_series = bool_series[
                bool_series.index.isin(result[result.isna() | (result == True)].index)
            ]
            result.loc[result.index.isin(bool_series.index)] = bool_series

    return result



from shapely.geometry import Point
import geopandas as gpd
from pathlib import Path

from hydamo_validation import logic_functions

try:
    from .config import DATA_DIR
except ImportError:
    from config import DATA_DIR

dataset_gpkg = DATA_DIR / "tasks" / "test_wrij" / "datasets" / "HyDAMO.gpkg"
gdf = gpd.read_file(dataset_gpkg, layer="Stuw")


def test_LE():
    _gdf = gdf.loc[gdf["kruinbreedte"].notna()].copy()
    expected_result = [False, True, True]
    result = logic_functions.LE(_gdf, left="kruinbreedte", right=3).to_list()
    assert result == expected_result


def test_LT():
    _gdf = gdf.loc[gdf["kruinbreedte"].notna()].copy()
    expected_result = [False, True, False]
    result = logic_functions.LT(_gdf, left="kruinbreedte", right=3).to_list()
    assert result == expected_result


def test_GE():
    _gdf = gdf.loc[gdf["kruinbreedte"].notna()].copy()
    expected_result = [True, False, True]
    result = logic_functions.GE(_gdf, left="kruinbreedte", right=3).to_list()
    assert result == expected_result


def test_GT():
    _gdf = gdf.loc[gdf["kruinbreedte"].notna()].copy()
    expected_result = [True, False, False]
    result = logic_functions.GT(_gdf, left="kruinbreedte", right=3).to_list()
    assert result == expected_result


def test_EQ():
    _gdf = gdf.loc[gdf["kruinbreedte"].notna()].copy()
    expected_result = [False, False, True]
    result = logic_functions.EQ(_gdf, left="kruinbreedte", right=3).to_list()
    assert result == expected_result


def test_BE():
    _gdf = gdf.loc[gdf["kruinbreedte"].notna()].copy()
    expected_result = [False, False, True]
    result = logic_functions.BE(
        _gdf, parameter="kruinbreedte", min=0.1, max=3.2
    ).to_list()
    assert result == expected_result


def test_BE_inclusive():
    _gdf = gdf.loc[gdf["kruinbreedte"].notna()].copy()
    expected_result = [False, True, True]
    result = logic_functions.BE(
        _gdf, parameter="kruinbreedte", min=0.1, max=3.2, inclusive=True
    ).to_list()
    assert result == expected_result


def test_ISIN():
    _gdf = gdf.loc[gdf["soortstuw"].notna()].copy()
    expected_result = [True, False, True]
    result = logic_functions.ISIN(
        _gdf, parameter="soortstuw", array=["overlaat"]
    ).to_list()
    assert result == expected_result


def test_NOTIN():
    _gdf = gdf.loc[gdf["soortstuw"].notna()].copy()
    expected_result = [False, True, False]
    result = logic_functions.NOTIN(
        _gdf, parameter="soortstuw", array=["overlaat"]
    ).to_list()
    assert result == expected_result



"""
Test voor datamodel.py
"""

from hydamo_validation.datamodel import HyDAMO
from pathlib import Path
import geopandas as gpd
from shapely.geometry import Point

try:
    from .config import DATA_DIR
except ImportError:
    from config import DATA_DIR

hydamo_version = "2.2"
object_layers = [
    "admingrenswaterschap",
    "afsluitmiddel",
    "afvoergebiedaanvoergebied",
    "aquaduct",
    "beheergrenswaterschap",
    "bijzonderhydraulischobject",
    "bodemval",
    "brug",
    "doorstroomopening",
    "duikersifonhevel",
    "gemaal",
    "grondwaterinfolijn",
    "grondwaterinfopunt",
    "grondwaterkoppellijn",
    "grondwaterkoppelpunt",
    "hydrologischerandvoorwaarde",
    "hydroobject",
    "hydroobject_normgp",
    "kunstwerkopening",
    "lateraleknoop",
    "meetlocatie",
    "meetwaardeactiewaarde",
    "normgeparamprofiel",
    "normgeparamprofielwaarde",
    "peilafwijkinggebied",
    "peilbesluitgebied",
    "peilgebiedpraktijk",
    "peilgebiedvigerend",
    "pomp",
    "profielgroep",
    "profiellijn",
    "profielpunt",
    "regelmiddel",
    "reglementgrenswaterschap",
    "ruwheidprofiel",
    "streefpeil",
    "sturing",
    "stuw",
    "vispassage",
    "vispassagevlak",
    "vuilvang",
    "zandvang",
]

ignored_layers = [
    "afvoeraanvoergebied",
    "imwa_geoobject",
    "leggerwatersysteem",
    "leggerwaterveiligheid",
    "waterbeheergebied",
]

dataset_gpkg = DATA_DIR / "tasks" / "test_wrij" / "datasets" / "HyDAMO.gpkg"
hydroobject_gdf = gpd.read_file(dataset_gpkg, layer="Hydroobject")
hydroobject_gdf.rename(
    columns={"ruwheidswaardehoog": "ruwheidhoog", "ruwheidswaardelaag": "ruwheidlaag"},
    inplace=True,
)

stuw_gdf = gpd.read_file(dataset_gpkg, layer="Stuw")

exports_dir = Path(__file__).parent / "exports"
exports_dir.mkdir(exist_ok=True)

datamodel = HyDAMO(version=hydamo_version)


def test_version():
    assert datamodel.version == hydamo_version


def test_layers():
    assert datamodel.layers == object_layers


def test_ignored_layers():
    assert datamodel.ignored_layers == ignored_layers


def test_setting_data():
    datamodel.set_data(hydroobject_gdf, "hydroobject")
    assert not datamodel.hydroobject.empty


def test_typeerror_data():
    gdf = hydroobject_gdf.copy()
    gdf.loc[0, "geometry"] = Point(0, 0)
    try:
        datamodel.set_data(gdf, "hydroobject")
        assert False
    except TypeError:
        assert True


def test_keyerror_missing_column():
    gdf = hydroobject_gdf.copy()
    gdf.drop("categorieoppwaterlichaam", axis=1, inplace=True)
    try:
        datamodel.hydroobject._check_columns(gdf)
        assert False
    except KeyError:
        assert True


def test_snapping_data():
    datamodel.set_data(hydroobject_gdf, "hydroobject", index_col=None)
    datamodel.set_data(stuw_gdf, "stuw", index_col=None)
    datamodel.stuw.snap_to_branch(datamodel.hydroobject, snap_method="overall")
    assert "branch_id" in datamodel.stuw.columns


def test_exporting_data(tmp_path):
    datamodel.set_data(hydroobject_gdf, "hydroobject", index_col=None)
    result = tmp_path.joinpath("datamodel_no_schema.gpkg")
    datamodel.to_geopackage(result)
    assert result.exists()
    result = tmp_path.joinpath("datamodel_schema.gpkg")
    datamodel.to_geopackage(result, use_schema=False)
    assert result.exists()



# %%
import geopandas as gpd
from hydamo_validation import general_functions
from hydamo_validation.datamodel import HyDAMO

try:
    from .config import DATA_DIR
except ImportError:
    from config import DATA_DIR

ahn_path = DATA_DIR / "dtm"

general_functions._set_coverage("ahn", ahn_path)
hydamo_version = "2.2"

dataset_gpkg = DATA_DIR / "tasks" / "test_wrij" / "datasets" / "HyDAMO.gpkg"
gdf = gpd.read_file(dataset_gpkg, layer="Stuw")
opening_gdf = gpd.read_file(dataset_gpkg, layer="Kunstwerkopening")

datamodel = HyDAMO(version=hydamo_version)


def test_coverage_not_found():
    try:
        general_functions._set_coverage("ahn", "test")
        assert False
    except FileNotFoundError:
        assert True


def test_sum_function_int():
    _gdf = gdf.loc[gdf["kruinbreedte"].notna()].copy()
    expected_result = [13.200000047683716, 10.1, 13.0]
    result = general_functions.sum(_gdf, ["kruinbreedte", 10]).to_list()
    assert result == expected_result


def test_sum_function_attr():
    _gdf = gdf.loc[gdf["kruinbreedte"].notna()].copy()
    _gdf["add"] = 10
    expected_result = [13.200000047683716, 10.1, 13.0]
    result = general_functions.sum(_gdf, ["kruinbreedte", "add"]).to_list()
    assert result == expected_result


def test_difference_function_int():
    _gdf = gdf.loc[gdf["kruinbreedte"].notna()].copy()
    expected_result = [-6.799999952316284, -9.9, -7.0]
    result = general_functions.difference(_gdf, left="kruinbreedte", right=10).to_list()
    assert result == expected_result


def test_difference_function_int_abs():
    _gdf = gdf.loc[gdf["kruinbreedte"].notna()].copy()
    expected_result = [6.799999952316284, 9.9, 7.0]
    result = general_functions.difference(
        _gdf, left="kruinbreedte", right=10, absolute=True
    ).to_list()
    assert result == expected_result


def test_devide_function_attr():
    _gdf = gdf.loc[gdf["kruinbreedte"].notna()].copy()
    _gdf["devide"] = 10
    expected_result = [0.3200000047683716, 0.01, 0.3]
    result = general_functions.divide(
        _gdf, left="kruinbreedte", right="devide"
    ).to_list()
    assert result == expected_result


def test_devide_function_int():
    _gdf = gdf.loc[gdf["kruinbreedte"].notna()].copy()
    expected_result = [0.3200000047683716, 0.01, 0.3]
    result = general_functions.divide(_gdf, left="kruinbreedte", right=10).to_list()
    assert result == expected_result


def test_difference_function_attr():
    _gdf = gdf.loc[gdf["kruinbreedte"].notna()].copy()
    _gdf["subtract"] = 10
    expected_result = [-6.799999952316284, -9.9, -7.0]
    result = general_functions.difference(
        _gdf, left="kruinbreedte", right="subtract"
    ).to_list()
    assert result == expected_result


def test_buffer_function_int():
    radius = 30
    percentile = 95
    expected_result = [32.17, 32.1, 35.12, 32.1, 32.23]
    result = general_functions.buffer(gdf, radius, percentile, coverage="ahn").to_list()
    assert result == expected_result


def test_buffer_function_attr():
    radius = "buffer_radius"
    percentile = 95
    _gdf = gdf.copy()
    _gdf.loc[:, "buffer_radius"] = _gdf.loc[:, "kruinbreedte"] + 10
    expected_result = [31.26, 32.1, 34.78]
    result = general_functions.buffer(
        _gdf.loc[_gdf["buffer_radius"].notna()], radius, percentile, coverage="ahn"
    ).to_list()
    assert result == expected_result


def test_buffer_function_fill():
    radius = 2
    percentile = 95
    expected_result = [50.0, 31.34, 50.0, 31.34, 50.0]
    result = general_functions.buffer(
        gdf, radius, percentile, coverage="ahn", fill_value=50
    ).to_list()
    assert result == expected_result


def test_object_relation_count():
    expected_result = [0.0, 1.0, 0.0, 0.0, 0.0]
    result = general_functions.object_relation(
        gdf,
        related_gdf=opening_gdf,
        code_relation="stuwid",
        statistic="count",
        related_parameter=None,
        fill_value=0,
    ).to_list()
    assert result == expected_result


def test_object_relation_majority():
    expected_result = [0, "Onbekend", 0, 0, 0]
    result = general_functions.object_relation(
        gdf,
        related_gdf=opening_gdf,
        code_relation="stuwid",
        statistic="majority",
        related_parameter="vormopening",
        fill_value=0,
    ).to_list()
    assert result == expected_result


def test_multiply():
    _gdf = gpd.GeoDataFrame(
        data={"left": [1, 2, 3], "right": [4, 5, 6], "geometry": [None, None, None]}
    )
    expected_result = [4, 10, 18]
    result = general_functions.multiply(_gdf, left="left", right="right").to_list()
    assert result == expected_result


# %%
